#!/usr/bin/perl -w
use strict;
use porter;
use Control;
use Backtrack;

my $loss = shift or die;

# example sequence labeling search

my %lists = ();

my $N = 0;
my @X = (); 
my @L = ();   # cache list info
my @Y = ();
my %l = ();

my $ON = 0; my $OL = 0;
my $NI = 0; my $NP = 0; my $NT = 0;
my $ACEM = 0; my $ACET = 0;

my %allwords = ();
=pod
print STDERR "Reading lists...";

open F, "ner.wordsicareabout" or die;
while (<F>) { chomp; $allwords{$_} = 1; }
close F or die;

open L, "ls ../pp/resources/list-* |" or die;
my $lnum = 0;
while (my $fn = <L>) {
    chomp $fn;
    print STDERR "$fn...";
    open F, $fn or die;
    while (<F>) {
        chomp;
        my $w = $_; my $v = $lnum;
        if (/^(.+)[\s]+(.+)$/) { $w = $1; $v = "$lnum:$2"; }
        if (exists $allwords{$w}) {
            $lists{$w}{$v} = 1;
        }
    }
    close F or die;
    $lnum++;
}
close L;
print STDERR "\n";
%allwords = ();
=cut

if ($loss eq 'hamming') {
    ControlSearn(
		 \&Load,
		 \&Num,
		 \&Example,
		 \&Actions,
		 \&Features,
		 \&Cost,
		 \&Step,
		 \&Loss,
		 \&OverallLoss);
} elsif ($loss eq 'f') {
    ControlSearn(
		 \&Load,
		 \&Num,
		 \&Example,
		 \&Actions,
		 \&Features,
		 \&Cost,
		 \&Step,
		 \&LossF,
		 \&OverallLossF);
} else {
    ControlSearn(
		 \&Load,
		 \&Num,
		 \&Example,
		 \&Actions,
		 \&Features,
		 \&Cost,
		 \&Step,
		 \&LossACE,
		 \&OverallLossACE);
}

sub Load {
    my ($fn) = @_;

    $N = 0; %l = (); @X = (); @Y = (); @L = ();
    open F, $fn or die;
    while (<F>) {
        chomp;
        my @xy = split;
        for (my $i=0; $i<@xy; $i++) {
            $xy[$i] =~ m/^(.+)_([^_]+)$/;
            $X[$N][$i] = $1;
            $Y[$N][$i] = $2;
            $l{$2} = 1;
        }
        $N++;
    }
    close F or die;

    $ON = 0; $OL = 0; $NI = 0; $NP = 0; $NT = 0; $ACEM = 0; $ACET = 0;
}

sub Num {
    return $N;
}

sub Example {
    my ($n) = @_;
    @L = ();
    for (my $i=0; $i<@{$X[$n]}; $i++) {
        @{$L[$i]} = ();
        my $w = lc($X[$n][$i]);
        if (exists $lists{$w}) {
            foreach my $l (keys %{$lists{$w}}) {
                push @{$L[$i]}, $l;
            }
        }
    }
    return ($n);
}

sub Actions {
    my ($n,@labs) = @_;
    my $m = scalar @{$X[$n]};
    if (scalar @labs >= $m) { return (); }
    else { return (sort keys %l); }
}

sub Features {
    my ($n,@labs) = @_;
    my %f = ();
    compute_features(\@{$X[$n]}, \@labs, \%f, 1, "");
    return (%f);
}

sub ActionFeatures {
    my ($n, @labs) = @_;

    my %f = ();
    my @l = ();
    for (my $i=0; $i<@labs; $i++) {
        compute_features(\@{$X[$n]}, \@l, \%f, 1, "$labs[$i]_");
        $l[$i] = $labs[$i];
    }

    return (%f);
}

sub Step {
    my ($action, @state) = @_;
    $state[@state] = $action;
    return @state;
}

sub Cost {
    my ($lab, $n, @labs) = @_;
    if ($lab eq $Y[$n][scalar @labs]) { return 0; }
    else { return 1; }
}

sub Loss {
    my ($n, @labs) = @_;
    my $l = 0;

    for (my $i=0; $i<@labs; $i++) {
        if ($labs[$i] ne $Y[$n][$i]) { $l++; $OL++; }
        $ON++;
    }

    LossACE($n,@labs);
    LossF($n,@labs);

    return $l;
}

sub OverallLoss {
    my $l = 0;
    if ($ON > 0) { $l = $OL/$ON; }

    my $overallAce = OverallLossACE();
    my $overallF   = OverallLossF();

    $ON = 0; $OL = 0; $NI = 0; $NP = 0; $NT = 0; $ACEM = 0; $ACET = 0;

    print STDERR " {ACE=$overallAce F=$overallF} ";

    return $l;
}

sub LossF {
    my ($n, @labs) = @_;

    my %c1 = chunksof(@labs);
    my %c2 = chunksof(@{$Y[$n]});

    my $np += scalar keys %c1;
    my $nt += scalar keys %c2;
    my $ni = 0;
    foreach my $c (keys %c1) {
        if (not exists $c2{$c}) { next; }
        $ni++;
    }

    $NP += $np; $NT += $nt; $NI += $ni;

    return (1-computef(0,$np,$nt,$ni));
}

sub OverallLossF {
    my $f = 1-computef(1,$NP,$NT,$NI);
    $ON = 0; $OL = 0; $NI = 0; $NP = 0; $NT = 0; $ACEM = 0; $ACET = 0;
    return $f;
}

sub LossACE {
    my ($n, @labs) = @_;

    my %c1 = chunksof(@labs);
    my %c2 = chunksof(@{$Y[$n]});

    my $maxAce  = ace_of($n, \%c2, \%c2);
    my $thisAce = ace_of($n, \%c2, \%c1);

    $ACEM += $maxAce;
    $ACET += $thisAce;

    if ($maxAce == 0) { return 0; }
    return (1 - ($thisAce / $maxAce));
}

sub OverallLossACE {
    my $l = ( $ACEM > 0) ? (1 - ($ACET / $ACEM)) : 0;
    $ON = 0; $OL = 0; $NI = 0; $NP = 0; $NT = 0; $ACEM = 0; $ACET = 0;
    return $l;
}

sub ace_of {
    my ($n,$t,$h) = @_;

    my $ace = 0;

    my %match = ();
    foreach my $x (keys %$h) {
        my ($st,$en,$tag) = split /[\s]+/, $x;
        my $len1 = string_length($n, $st ,$en , $st , $en );

        foreach my $y (keys %$t) {
            my ($st2,$en2,$tag2) = split /[\s]+/, $y;
            if (overlaps($st,$en,$st2,$en2)) {
                my $len = string_length($n,$st,$en,$st2,$en2);
                my $v = 1;
                my $len2 = string_length($n, $st2,$en2, $st2, $en2);
                my $max_len = ($len1 > $len2) ? $len1 : $len2;
                $v *= ($len / $max_len);
                if ($tag ne $tag2) { $v *= 0.5; }
                $match{$x . "\t" . $y} = $v;
            }
        }
    }

    my %usedX = (); my %usedY = (); my %allX = (); my %allY = ();
    foreach my $xy (sort { $match{$b} <=> $match{$a} } keys %match) {
        my ($x,$y) = split /\t/, $xy;
        $allX{$x} = 1; $allY{$y} = 1;
        if ((exists $usedX{$x}) || (exists $usedY{$y})) { next; }
        $ace += $match{$xy};
        $usedX{$x} = 1;
        $usedY{$y} = 1;
    }

    # now, handle false alarms
    foreach my $x (keys %allX) {
        if (exists $usedX{$x}) { next; }
        $ace -= 0.5;
    }

    # and misses
    foreach my $y (keys %allY) {
        if (exists $usedY{$y}) { next; }
        $ace -= 0.5;
    }
        

#        my $longest = 0;
#        my $longval = "";
#        foreach my $y (keys %$t) {
#            my ($st2,$en2,$tag2) = split /[\s]+/, $y;
#            if (overlaps($st,$en,$st2,$en2)) {
#                my $len = string_length($n,$st,$en,$st2,$en2);
#                if ($len >= $longest) {
#                    $longest = $len;
#                    $longval = $y;
#                }
#            }
#        }
#        if ($longest <= 0) { next; }
#        my ($st2,$en2,$tag2) = split /[\s]+/, $longval;
#
#        $usedT{$longval}{LEN} = $len;
#        $usedT{$longval}{VAL} = $v;
#
#        $ace += $v;
#    }

    return $ace;
}

sub overlaps {
    my ($s,$e,$s2,$e2) = @_;
    my $ss = ($s < $s2) ? $s2 : $s;
    my $ee = ($e > $e2) ? $e2 : $e;
    if ($ss <= $ee-1) { return 1; }
    return 0;
}

sub string_length {
    my ($n,$s,$e,$s2,$e2) = @_;
    my $ss = ($s < $s2) ? $s2 : $s;
    my $ee = ($e > $e2) ? $e2 : $e;
    my $len = -1;
    for (my $i=$ss; $i<$ee; $i++) {
        $len += 1 + length($X[$n][$i]);
    }
    if ($len < 0) { return 0; }
    return $len;
}

sub chunksof {
    my @l = @_;
    my $i = 0;
    my %c = ();
    while ($i < @l) {
        if ($l[$i] =~ /^B-(.+)$/) {
            my $lab = $1;
            if ($lab eq 'O') { $i++; next; }
            my $j = $i+1;
            while ($j < @l) {
                if ($l[$j] eq "I-$lab") { $j++; }
                else { last; }
            }
            $c{"$i $j $lab"} = 1;
            $i = $j;
        } else {
            $i++;
        }
    }
    return (%c);
}

sub computef {
    my ($last,$np,$nt,$ni) = @_;
    if ($last) { print STDERR "[$np $nt $ni]"; }
    if (($np == 0) && ($nt == 0)) { return 1; }
    if ($ni == 0) { return 0; }
    my $p = $ni / $np;
    my $r = $ni / $nt;
    if ($last) { print STDERR "[$p $r]"; }
    return (2*$p*$r / ($p + $r));
}











# features

sub compute_features {
    my ($x, $lab, $f, $val, $str) = @_;

    my $i = (scalar @{$lab});
    my $M = (scalar @{$x});

    my $pl = "<s>"; my $pl2 = "<s>";
    if ($i > 0) { $pl = $lab->[$i-1]; }
    if ($i > 1) { $pl = $lab->[$i-2]; }
    $f->{$str . "pl1=" . $pl} += $val;
    $f->{$str . "pl2=" . $pl2} += $val;
    $f->{$str . "pl3=" . $pl . "//" . $pl2} += $val;

    for (my $w=-2; $w<=2; $w++) {
        if ($i+$w < 0) { next; }
        if ($i+$w >= $M) { next; }

        compute_features_at($x,$str . "AT$w",$i+$w,$f,$val);
        compute_features_at($x,$str . "$pl-AT$w",$i+$w,$f,$val);
    }

}

sub compute_features_at {
    my ($x,$str,$i,$f,$val) = @_;

    my $w = $x->[$i];
    my $l = lc($w);

    my $case = $w;
    $case =~ s/[A-Z]/A/g;
    $case =~ s/[a-z]/a/g;
    $case =~ s/[0-9]/0/g;
    $case =~ s/[^Aa0\.]/x/g;
    $case =~ s/AA+/A\+/g;
    $case =~ s/aa+/a\+/g;
    $case =~ s/00+/0\+/g;
    $case =~ s/xx+/x\+/g;

    my $case2 = $case;
    $case2 =~ s/\+//g;

    $f->{"$str:" . "WORD-" . $w} += $val;
    $f->{"$str:" . "LC-"   . lc($w)} += $val;
    $f->{"$str:" . "STEM-" . stem($l)} += $val;
    $f->{"$str:" . "PRE1-" . prefix($l,1)} += $val;
    $f->{"$str:" . "PRE2-" . prefix($l,2)} += $val;
    $f->{"$str:" . "PRE3-" . prefix($l,3)} += $val;
    $f->{"$str:" . "SUF1-" . suffix($l,1)} += $val;
    $f->{"$str:" . "SUF2-" . suffix($l,2)} += $val;
    $f->{"$str:" . "SUF3-" . suffix($l,3)} += $val;
    $f->{"$str:" . "CASE-" . $case} += $val;
    $f->{"$str:" . "CAS+-" . $case2} += $val;

    foreach my $l (@{$L[$i]}) {
        $f->{"$str:list" . $l}++;
    }

#    if (exists $lists{lc($w)}) {
#        foreach my $l (keys %{$lists{lc($w)}}) {
#            $f->{"$str:list:" . $l} = 1;
#        }
#    }
}

sub prefix {
    my ($w,$n) = @_;
    my $l = length($w);
    if ($n >= $l) { return $w; }
    return substr($w,0,$n);
}
                                                                                                                                                                                                              
sub suffix {
    my ($w,$n) = @_;
    my $l = length($w);
    if ($n >= $l) { return $w; }
    return substr($w,$l-$n,$n);
}
 
